@@ -38,7 +38,7 @@ from paver.easy import Bunch, options, task, sh
 #-----------------------------------
 
 # Path to the release notes
-RELEASE_NOTES = 'doc/source/release/1.24.4-notes.rst'
+RELEASE_NOTES = 'doc/source/release/1.25.0-notes.rst'
 
 
 #-------------------------------------------------------
@@ -184,7 +184,7 @@ def write_release_task(options, filename='README'):
     ----------
     options :
         Set by ``task`` decorator.
-    filename : string
+    filename : str
         Filename of the modified notes. The file is written
         in the release directory.
 
